// SPDX-License-Identifier: GPL-2.0
/* Copyright (C) 2009~2015 Lantiq Deutschland GmbH
 * Copyright (C) 2016~2018 Intel Corporation.
 */

#define CONFIG_USERSPACE_LINK_NOTIFICATION

#include <linux/version.h>
#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/interrupt.h>
#include <linux/kernel.h> /* printk() */
#include <linux/types.h> /* size_t */
#include <linux/etherdevice.h>
#include <linux/ethtool.h>
#include <linux/proc_fs.h>
#include <linux/etherdevice.h> /* eth_type_trans */
#include <linux/delay.h>
#include <linux/init.h>
#include <linux/clk.h>
#include <linux/of_net.h>
#include <linux/of_mdio.h>
#include <linux/of_irq.h>
#include <linux/of_device.h>
#include <linux/env.h>

#ifdef CONFIG_USERSPACE_LINK_NOTIFICATION
#include <net/genetlink.h>
#endif

#include <lantiq.h>
#include <net/datapath_api.h>
#include <net/lantiq_cbm_api.h>
#include <net/switch_api/lantiq_gsw_api.h>
#include <net/switch_api/gsw_dev.h>
#include "xrx500_phy_fw.h"
#include "ltq_eth_drv_xrx500.h"
#include "datapath/datapath.h"

#if IS_ENABLED(CONFIG_AVM_PA)
#include <avm/pa/avm_pa.h>
#endif

#define LTQ_ETH_MAX_DATA_LEN 9000

#define DRV_MODULE_NAME "lantiq_eth_drv_xrx500"
#define DRV_MODULE_VERSION "1.1"

/* length of time before we decide the hardware is borked,
 * and dev->eth_tx_timeout() should be called to fix the problem
 */
#define LTQ_TX_TIMEOUT (10 * HZ)
#define MDIO_PHY_ADDR_MASK 0x001f
#define MDIO_PHY_LINK_DOWN 0x4000
#define MDIO_PHY_LINK_UP 0x2000

#define MDIO_PHY_FDUP_EN 0x0600
#define MDIO_PHY_FDUP_DIS 0x0200

#define LTQ_ETH_NUM_INTERRUPTS 5

/* Init of the network device */
static int ltq_eth_init(struct net_device *dev);
/* Start the network device interface queue */
static int ltq_eth_open(struct net_device *dev);
/* Stop the network device interface queue */
static int ltq_eth_stop(struct net_device *dev);
/* Uninit the network device interface queue */
static void ltq_eth_uninit(struct net_device *dev);
/* Transmit packet from Tx Queue to MAC */
static int ltq_start_xmit(struct sk_buff *skb, struct net_device *dev);
/* Hardware specific IOCTL's  */
static int ltq_ioctl(struct net_device *dev, struct ifreq *ifr, int cmd);
/* Get the network device statistics */
static struct rtnl_link_stats64 *ltq_get_stats(struct net_device *dev,
					       struct rtnl_link_stats64 *storage);
/* change MTU values */
static int ltq_change_mtu(struct net_device *dev, int new_mtu);
/*  Set mac address*/
static int ltq_set_mac_address(struct net_device *dev, void *p);
/* Transmit timeout*/
static void ltq_tx_timeout(struct net_device *dev);

/* interface change event handler */
static int phy_netdevice_event(struct notifier_block *nb, unsigned long action,
			       void *ptr);
static int xrx500_mdio_probe(struct net_device *dev, struct xrx500_port *port);
static int ltq_gsw_pmac_init(void);

/**
 *  Datapath directpath functions
 */
static int32_t dp_fp_stop_tx(struct net_device *);
static int32_t dp_fp_restart_tx(struct net_device *);
static int32_t dp_fp_rx(struct net_device *, struct net_device *,
			struct sk_buff *, int32_t);

#ifdef CONFIG_XRX500_ETH_DRV_COC_SUPPORT
static int32_t dp_fp_coc_confirm(enum ltq_cpufreq_state new_state,
				 enum ltq_cpufreq_state old_state, u32 flags);
DECLARE_BITMAP(g_ltq_eth_intr_type, LTQ_ETH_NUM_INTERRUPTS);
static u32 g_ltq_eth_gswl_irq;
static struct tasklet_struct gswl_tasklet;

DECLARE_BITMAP(g_ltq_pae_intr_type, LTQ_ETH_NUM_INTERRUPTS);
static u32 g_ltq_pae_irq;
static struct tasklet_struct pae_tasklet;
static enum ltq_cpufreq_state g_ltq_eth_drv_coc_state = LTQ_CPUFREQ_PS_D0;
#endif

static struct xrx500_hw xrx500_hw;

static char wan_iface[IFNAMSIZ] = "eth1";
static char *macaddr0;
static char *macaddr1;
static int multimac = 1;

static struct module g_ltq_eth_module[NUM_ETH_INF];
static u32 g_rx_csum_offload;
static u32 g_eth_switch_mode;
static struct ltq_net_soc_data g_soc_data;

/*  AVMNET stat functions
 */
static void avmnet_get_ethtool_stats(struct net_device *netdev,
				     struct ethtool_stats *stats, u64 *data);
static void avmnet_get_strings(struct net_device *netdev,
			       u32 stringset, u8 *data);
static int avmnet_get_sset_count(struct net_device *netdev, int string_set);

static const struct net_device_ops ltq_eth_drv_ops = {
	.ndo_init = ltq_eth_init,
	.ndo_open = ltq_eth_open,
	.ndo_stop = ltq_eth_stop,
	.ndo_uninit = ltq_eth_uninit,
	.ndo_start_xmit = ltq_start_xmit,
	.ndo_set_mac_address = ltq_set_mac_address,
	.ndo_change_mtu = ltq_change_mtu,
	.ndo_get_stats64 = ltq_get_stats,
	.ndo_do_ioctl = ltq_ioctl,
	.ndo_tx_timeout = ltq_tx_timeout,
};

static struct notifier_block netdevice_notifier = {
	.notifier_call = phy_netdevice_event
};

void gsw_reg_set_val(struct core_ops *ops, u16 nRegAddr, u16 nData);
u16 gsw_reg_get_val(struct core_ops *ops, u16 nRegAddr);

#undef DUMP_PACKET

#ifdef DUMP_PACKET
/**
 * \brief dump skb data
 * \param[in] len length of the data buffer
 * \param[in] p_data Pointer to data to dump
 *
 * \return void No Value
 */
static inline void dump_skb(u32 len, char *p_data)
{
	int i;

	for (i = 0; i < len; i++) {
		pr_info("%2.2x ", (u8)(p_data[i]));

		if (i % 16 == 15)
			pr_info("\n");
	}

	pr_info("\n");
}
#endif

/* Get the driver information, used by ethtool_ops  */
static void get_drvinfo(struct net_device *dev, struct ethtool_drvinfo *info)
{
	/* driver driver short name (Max 32 characters) */
	strcpy(info->driver, DRV_MODULE_NAME);
	/* driver version(Max 32 characters) */
	strcpy(info->version, DRV_MODULE_VERSION);
}

/* Get the network device settings  */
static int get_settings(struct net_device *dev, struct ethtool_cmd *cmd)
{
	struct ltq_eth_priv *priv;
	int i;

	priv = netdev_priv(dev);

	pr_debug("in ethtool get setting for num %d\n", priv->num_port);
	for (i = 0; i < priv->num_port; i++)
		if (priv->port[i].phydev)
			return phy_ethtool_gset(priv->port[i].phydev, cmd);

	return -ENODEV;
}

/* Set the network device settings */
static int set_settings(struct net_device *dev, struct ethtool_cmd *cmd)
{
	struct ltq_eth_priv *priv;
	int i;

	priv = netdev_priv(dev);
	pr_info("%s: called\n", __func__);

	for (i = 0; i < priv->num_port; i++)
		if (priv->port[i].phydev)
			return phy_ethtool_sset(priv->port[i].phydev, cmd);

	return -ENODEV;
}

#if IS_ENABLED(CONFIG_LTQ_DATAPATH_PTP1588)
static int get_tsinfo(struct net_device *dev, struct ethtool_ts_info *ts_info)
{
	struct ltq_eth_priv *priv;
	struct mac_ops *ops;

	priv = netdev_priv(dev);

	ops = gsw_get_mac_ops(0, (priv->dp_port_id - 2));
	if (ops)
		ops->mac_get_ts_info(ops, ts_info);

	return 0;
}
#endif

/* Reset the device */
static int nway_reset(struct net_device *dev)
{
	struct ltq_eth_priv *priv = netdev_priv(dev);
	int i;

	for (i = 0; i < priv->num_port; i++)
		if (priv->port[i].phydev)
			genphy_restart_aneg(priv->port[i].phydev);
	return 0;
}

#define MAC_CTRL_4 0x907
#define MAC_CTRL_4_LPIEN_SHIFT 7
#define MAC_CTRL_4_GWAIT_SHIFT 8
/*
 * AVM: we use edata.tx_lpi_timer upper bits for 1G-delay
 * 	we use edata.tx_lpi_timer lower bits for 100M-delay
 */
static void ltq_eee_mac_set(struct ltq_eth_priv *priv, struct ethtool_eee *edata)
{
	struct core_ops *ops;
	u16 nRegAddr;
	u32 delay_100m = (edata->tx_lpi_timer & 0x7F);
	u32 delay_1g = ((edata->tx_lpi_timer >> 16) & 0x7F);
	u32 lpi_enabled = !!(edata->tx_lpi_enabled);

	ops = gsw_get_swcore_ops(!!(priv->wan));

	if (!ops) {
		pr_err("%s: Open SWAPI device FAILED!\n", __func__);
	}

	if (priv->wan)
		nRegAddr = MAC_CTRL_4;
	else
		nRegAddr = MAC_CTRL_4 + (0xc * (priv->id));

	gsw_reg_set_val(ops, nRegAddr, (delay_1g << MAC_CTRL_4_GWAIT_SHIFT) |
		       (lpi_enabled << MAC_CTRL_4_LPIEN_SHIFT) | delay_100m);
}

/*
 * AVM: we use edata.tx_lpi_timer upper bits for 1G-delay
 * 	we use edata.tx_lpi_timer lower bits for 100M-delay
 */
static void ltq_eee_mac_get(struct ltq_eth_priv *priv, struct ethtool_eee *edata)
{
	struct core_ops *ops;
	u16 nRegAddr;
	u32 nData;
	u32 delay_100m;
	u32 delay_1g;
	u32 lpi_enabled;

	ops = gsw_get_swcore_ops(!!(priv->wan));

	if (!ops) {
		pr_err("%s: Open SWAPI device FAILED!\n", __func__);
		return;
	}

	if (priv->wan)
		nRegAddr = MAC_CTRL_4;
	else
		nRegAddr = MAC_CTRL_4 + (0xc * (priv->id));

	nData = gsw_reg_get_val(ops, nRegAddr);

	delay_1g = (nData >> MAC_CTRL_4_GWAIT_SHIFT) & 0x7F;
	delay_100m = nData & 0x7F;
	lpi_enabled = (nData >> MAC_CTRL_4_LPIEN_SHIFT) & 0x1;

	edata->tx_lpi_timer = (delay_1g << 16) | delay_100m;
	edata->tx_lpi_enabled = lpi_enabled;
}

static int ltq_get_eee(struct net_device *dev, struct ethtool_eee *edata)
{
	int i;
	struct ltq_eth_priv *priv = netdev_priv(dev);

	memset(edata, 0, sizeof(struct ethtool_eee));

	for (i = 0; i < priv->num_port; i++) {
		if (priv->port[i].phydev) {
			uint32_t mii_status;
			int res = phy_ethtool_get_eee(priv->port[i].phydev,
						      edata);

			mii_status = phy_read(priv->port[i].phydev, 0x18);

			edata->eee_active = !!(mii_status & (1 << 2));
			edata->eee_enabled = !!(priv->eee_enabled);
			ltq_eee_mac_get(priv, edata);
			pr_debug(
				"eee_status: %#x, active=%d, enabled=%d, (neg: readstat=%#x, supp=%#x, adv=%#x, p=%#x)\n",
				mii_status, edata->eee_active,
				edata->eee_enabled, res, edata->supported,
				edata->advertised, edata->lp_advertised);

			return res;
		}
	}

	return -ENODEV;
}

static unsigned int phy_restart_aneg_if_active(struct phy_device *phydev)
{
	unsigned int ctl;
	ctl = phy_read(phydev, MII_BMCR);

	if (ctl < 0)
		return ctl;

	if (!(ctl & BMCR_ANENABLE))
		return -EINVAL;

	if (ctl & BMCR_ISOLATE)
		return -EINVAL;

	ctl |= (BMCR_ANENABLE | BMCR_ANRESTART);

	ctl = phy_write(phydev, MII_BMCR, ctl);
	return ctl;
}

static int ltq_set_eee(struct net_device *dev, struct ethtool_eee *edata)
{
	int i;
	struct ltq_eth_priv *priv = netdev_priv(dev);

	for (i = 0; i < priv->num_port; i++) {
		if (priv->port[i].phydev) {
			int res;
			priv->eee_enabled = edata->eee_enabled;
			res = phy_ethtool_set_eee(priv->port[i].phydev, edata);
			ltq_eee_mac_set(priv, edata);
			phy_restart_aneg_if_active(priv->port[i].phydev);
			return res;
		}
	}
	return -ENODEV;
}

/* Structure of the ether tool operation  */
static const struct ethtool_ops ethtool_ops = {
	.get_drvinfo = get_drvinfo,
	.get_settings = get_settings,
	.set_settings = set_settings,
	.nway_reset = nway_reset,
	.get_link = ethtool_op_get_link,
	.get_link_ksettings = phy_ethtool_get_link_ksettings,
	.set_link_ksettings = phy_ethtool_set_link_ksettings,
	.get_eee = ltq_get_eee,
	.set_eee = ltq_set_eee,
#if IS_ENABLED(CONFIG_LTQ_DATAPATH_PTP1588)
	.get_ts_info = get_tsinfo,
#endif
	.get_sset_count = avmnet_get_sset_count,
	.get_ethtool_stats = avmnet_get_ethtool_stats,
	.get_strings = avmnet_get_strings,
};

/* open the network device interface*/
static int ltq_eth_open(struct net_device *dev)
{
	int ret;
	int flags = 1; /* flag 1 for enable */

	pr_debug("%s called for device %s\n", __func__, dev->name);
#ifdef CONFIG_GRX500_CBM
	/* Enable the p2p channel at CBM */
	if (!cbm_turn_on_DMA_p2p())
		pr_info("p2p channel turned ON !\n");
	else
		pr_info("p2p channel already ON !\n");
#endif

	ret = dp_rx_enable(dev, dev->name, flags);

	if (ret != DP_SUCCESS) {
		pr_err("%s: failed to enable for device: %s ret %d\n",
		       __func__, dev->name, ret);
		return -1;
	}

	return 0;
}

/* Stop the network device interface*/
static int ltq_eth_stop(struct net_device *dev)
{
	int ret;
	int flags = 0; /* flag 0 for disable */

	pr_debug("%s called for device %s\n", __func__, dev->name);
	ret = dp_rx_enable(dev, dev->name, flags);

	if (ret != DP_SUCCESS) {
		pr_err("%s: failed to disable for device: %s ret %d\n",
		       __func__, dev->name, ret);
		return -1;
	}
	/* netif_stop_queue(dev); */

	return 0;
}

/* Uninit the network device interface*/
static void ltq_eth_uninit(struct net_device *dev)
{
	struct ltq_eth_priv *priv;
	int ret;

	pr_debug("%s called for device %s\n", __func__, dev->name);
	priv = netdev_priv(dev);
	priv->dp_subif.subif = 0;
	priv->dp_subif.port_id = priv->dp_port_id;

	if (priv->dp_port_id == DP_FAILURE) {
		pr_debug("dp port id (%d) is invalid. ignore the deregister.\n",
			 priv->dp_port_id);
		return;
	}
	ret = dp_register_subif(priv->owner, dev, dev->name, &priv->dp_subif,
				DP_F_DEREGISTER);

	if (ret != DP_SUCCESS) {
		pr_err("%s: failed to close for device: %s ret %d\n",
		       __func__, dev->name, ret);
	}
}

/* Send the packet to netwrok rx queue */
static void eth_rx(struct net_device *dev, int len, struct sk_buff *skb)
{
	struct ltq_eth_priv *priv;

	priv = netdev_priv(dev);
	skb->dev = dev;
	skb->protocol = eth_type_trans(skb, dev);

	if (dev->features & NETIF_F_RXCSUM)
		skb->ip_summed = CHECKSUM_UNNECESSARY;

	pr_debug("passing to stack: protocol: %x\n", skb->protocol);
	netif_rx(skb);
	priv->stats.rx_packets++;
	priv->stats.rx_bytes += len;
}

static int32_t dp_fp_stop_tx(struct net_device *netif)
{
	return 0;
}

static int32_t dp_fp_restart_tx(struct net_device *netif)
{
	return 0;
}

static int32_t dp_fp_rx(struct net_device *rxif, struct net_device *txif,
			struct sk_buff *skb, int32_t len)
{
	/*skb_put(skb,len);*/
#ifdef DUMP_PACKET
	if (skb->data) {
		pr_ifo("raw data len:%d\n", len);
		dump_skb(len, (char *)skb->data);
	}

#endif

	if (skb) {
		/*Remove PMAC to DMA header */
		len -= (sizeof(ltq_pmac_header_t));
		skb_pull(skb, (sizeof(ltq_pmac_header_t)));
	} else {
		pr_err("%s: skb from DP is null !\n", __func__);
		goto rx_err_exit;
	}

	/* Pass it to the stack */
#ifdef DUMP_PACKET

	if (skb->data) {
		pr_info("data sent to stack.\n");
		dump_skb(len, (char *)skb->data);
	}

#endif

	if (rxif) {
		pr_debug("%s: rxed a packet from DP lib on interface %s\n",
			 __func__, rxif->name);

#if CONFIG_AVM_PA
		if (avm_pa_dev_pid_receive(AVM_PA_DEVINFO(rxif), skb) == AVM_PA_RX_STOLEN) {
			return 0;
		}
#endif

		eth_rx(rxif, len, skb);
	} else {
		pr_err("%s: error: rxed a packet from DP lib on interface %x\n",
		       __func__, (unsigned int)rxif);
		goto rx_err_exit;
	}

	return 0;
rx_err_exit:

	if (skb)
		dev_kfree_skb_any(skb);

	return -1;
}

/* Get the network device stats information */
static struct
	rtnl_link_stats64 *
	ltq_get_stats(struct net_device *dev,
		      struct rtnl_link_stats64 *storage)
{
#ifndef CONFIG_LTQ_DATAPATH_MIB
	struct ltq_eth_priv *priv = netdev_priv(dev);
	*storage = priv->stats;
#else
	dp_get_netif_stats(dev, NULL, storage, 0);
#endif
	return storage;
}

/* Trasmit timeout */
static void ltq_tx_timeout(struct net_device *dev)
{
}

/* Set the MAC address */
static int ltq_set_mac_address(struct net_device *dev, void *p)
{
	struct sockaddr *addr = p;
	struct ltq_eth_priv *priv = netdev_priv(dev);

	pr_debug("set_mac_addr called\n");
	if (netif_running(dev))
		return -EBUSY;

	if (!is_valid_ether_addr(addr->sa_data))
		return -EINVAL;

	memcpy(dev->dev_addr, addr->sa_data, dev->addr_len);

	if (priv->xgmac_id >= 0) {
		struct mac_ops *ops;

		ops = gsw_get_mac_ops(0, priv->xgmac_id);
		if (ops)
			ops->set_macaddr(ops, dev->dev_addr);
	}
	return 0;
}

void gsw_reg_set_bit(struct core_ops *ops, u16 nRegAddr, u16 nData)
{
	GSW_register_t reg_cfg;

	memset((void *)&reg_cfg, 0x00, sizeof(reg_cfg));
	reg_cfg.nRegAddr = nRegAddr;
	ops->gsw_common_ops.RegisterGet(ops, &reg_cfg);
	reg_cfg.nData |= nData;
	ops->gsw_common_ops.RegisterSet(ops, &reg_cfg);
}

void gsw_reg_clr_bit(struct core_ops *ops, u16 nRegAddr, u16 nData)
{
	GSW_register_t reg_cfg;

	memset((void *)&reg_cfg, 0x00, sizeof(reg_cfg));
	reg_cfg.nRegAddr = nRegAddr;
	ops->gsw_common_ops.RegisterGet(ops, &reg_cfg);
	reg_cfg.nData &= ~nData;
	ops->gsw_common_ops.RegisterSet(ops, &reg_cfg);
}

void gsw_reg_set_val(struct core_ops *ops, u16 nRegAddr, u16 nData)
{
	GSW_register_t reg_cfg;

	memset((void *)&reg_cfg, 0x00, sizeof(reg_cfg));
	reg_cfg.nRegAddr = nRegAddr;
	reg_cfg.nData = nData;
	ops->gsw_common_ops.RegisterSet(ops, &reg_cfg);
}

u16 gsw_reg_get_val(struct core_ops *ops, u16 nRegAddr)
{
	GSW_register_t reg_cfg;

	memset((void *)&reg_cfg, 0x00, sizeof(reg_cfg));
	reg_cfg.nRegAddr = nRegAddr;
	ops->gsw_common_ops.RegisterGet(ops, &reg_cfg);

	return reg_cfg.nData;
}

static int ltq_enable_gsw_l_jumbo(struct net_device *dev)
{
	struct ltq_eth_priv *priv = netdev_priv(dev);
	struct core_ops *ops;

	pr_info("%s called for id: %d\n", __func__, priv->id);

	/* Do the GSW-L configuration */
	pr_info("doing the GSW-L configuration for jumbo\n");
	ops = gsw_get_swcore_ops(0);

	if (!ops) {
		pr_err("%s: Open SWAPI device FAILED!\n", __func__);
		return -EIO;
	}

	/* Set the MAC control register 2 to enable Jumbo frames */
	gsw_reg_set_bit(ops, (0x8f9 + (0xc * (priv->id + 1))), 0x8);

	/* PMAC control register 2 to enable jumbo frames */
	pr_info("doing the PMAC configuration for jumbo at GSW-L\n");
	gsw_reg_set_bit(ops, 0xd05, 0x8);

	/* Set the jumbo frame length threshold */
	gsw_reg_set_val(ops, 0xd06, 9216);

	return 0;
}

static int ltq_disable_gsw_l_jumbo(struct net_device *dev)
{
	struct ltq_eth_priv *priv = netdev_priv(dev);
	struct core_ops *ops;

	pr_info("%s called for id: %d\n", __func__, priv->id);

	/* Do the GSW-L configuration */
	pr_info("doing the GSW-L configuration for jumbo\n");
	ops = gsw_get_swcore_ops(0);

	if (!ops) {
		pr_err("%s: Open SWAPI device FAILED!\n", __func__);
		return -EIO;
	}

	/* Set the MAC control register 2 to disable Jumbo frames */
	gsw_reg_clr_bit(ops, (0x8f9 + (0xc * (priv->id + 1))), 0x8);

	return 0;
}

static int ltq_enable_gsw_r_pmac_jumbo(struct net_device *dev)
{
	/* Do the GSW-R configuration */
	struct core_ops *ops = gsw_get_swcore_ops(1);

	if (!ops) {
		pr_err("%s: Open SWAPI device FAILED!\n", __func__);
		return -EIO;
	}

	/* PMAC control register 2 to enable jumbo frames */
	pr_info("doing the PMAC configuration for jumbo at GSW-R\n");
	gsw_reg_set_bit(ops, 0xd05, 0x8);

	/* Set the jumbo frame length threshold */
	gsw_reg_set_val(ops, 0xd06, 1662);

	return 0;
}

static int ltq_enable_gsw_r_jumbo(struct net_device *dev)
{
	/* Do the GSW-R configuration */
	struct core_ops *ops = gsw_get_swcore_ops(1);

	if (!ops) {
		pr_err("%s: Open SWAPI device FAILED!\n", __func__);
		return -EIO;
	}

	/* Set the MAC control register 2 to enable Jumbo frames */
	pr_info("doing the GSW-R configuration for jumbo\n");
	gsw_reg_set_bit(ops, 0x905, 0x8);

	return 0;
}

static int ltq_disable_gsw_r_jumbo(struct net_device *dev)
{
	/* Do the GSW-R configuration */
	struct core_ops *ops = gsw_get_swcore_ops(1);

	if (!ops) {
		pr_err("%s: Open SWAPI device FAILED!\n", __func__);
		return -EIO;
	}

	/* Set the MAC control register 2 to enable Jumbo frames */
	pr_info("doing the GSW-R configuration for jumbo\n");
	gsw_reg_clr_bit(ops, 0x905, 0x8);

	return 0;
}

/* Change the MTU value of the netwrok device interfaces */
static int ltq_change_mtu(struct net_device *dev, int new_mtu)
{
	struct ltq_eth_priv *priv = netdev_priv(dev);

	if (new_mtu < ETH_ZLEN || new_mtu > LTQ_ETH_MAX_DATA_LEN)
		return -EINVAL;

	/* if the MTU > 1500, do the jumbo config in switch */
	if (new_mtu > ETH_DATA_LEN && !(priv->jumbo_enabled)) {
		if (priv->wan) {
			if (ltq_enable_gsw_r_jumbo(dev) < 0)
				return -EIO;
		} else {
			if (ltq_enable_gsw_l_jumbo(dev) < 0)
				return -EIO;
		}

		ltq_enable_gsw_r_pmac_jumbo(dev);
		priv->jumbo_enabled = 1;
	} else if (new_mtu <= ETH_DATA_LEN && priv->jumbo_enabled) {
		if (priv->wan)
			ltq_disable_gsw_r_jumbo(dev);
		else
			ltq_disable_gsw_l_jumbo(dev);

		priv->jumbo_enabled = 0;
	}

	dev->mtu = new_mtu;
	return 0;
}

static int ltq_start_xmit(struct sk_buff *skb, struct net_device *dev)
{
	struct ltq_eth_priv *priv = netdev_priv(dev);
	int ret;
	int len;

	/* Call the Datapath Library's TX function */
	((struct dma_tx_desc_1 *)&skb->DW1)->field.ep = priv->dp_subif.port_id;
	((struct dma_tx_desc_0 *)&skb->DW0)->field.dest_sub_if_id =
		priv->dp_subif.subif;

	len = skb->len;

	if (dev->features & NETIF_F_HW_CSUM) {
		ret = dp_xmit(dev, &priv->dp_subif, skb, skb->len,
			      DP_TX_CAL_CHKSUM);
	} else {
		ret = dp_xmit(dev, &priv->dp_subif, skb, skb->len, 0);
	}
	if (!ret) {
		priv->stats.tx_packets++;
		priv->stats.tx_bytes += len;
	} else {
		priv->stats.tx_dropped++;
	}
	return 0;
}

/* Platform specific IOCTL's handler */
static int ltq_ioctl(struct net_device *dev, struct ifreq *ifr, int cmd)
{
	return -EOPNOTSUPP;
}

/* init of the network device */
static int ltq_eth_init(struct net_device *dev)
{
	int i;
	struct ltq_eth_priv *priv;
	int ret;

	priv = netdev_priv(dev);
	pr_debug("probing for number of ports = %d\n", priv->num_port);
	pr_debug("%s called for device %s\n", __func__, dev->name);

	for (i = 0; i < priv->num_port; i++) {
		if (of_phy_is_fixed_link(priv->port[i].phy_node))
			pr_debug("phy is fixed-link\n");
		else if (xrx500_mdio_probe(dev, &priv->port[i]))
			pr_warn("xrx500-mdio: probing phy of port %d failed\n",
				priv->port[i].num);
		dev->ethtool_ops = &ethtool_ops;
	}

#if IS_ENABLED(CONFIG_LTQ_DATAPATH_PTP1588)
	dev->ethtool_ops = &ethtool_ops;
#endif

	priv->dp_subif.subif = 0;
	priv->dp_subif.port_id = priv->dp_port_id;

	ret = dp_register_subif(priv->owner, dev, dev->name, &priv->dp_subif,
				0);

	if (ret != DP_SUCCESS) {
		pr_err("%s: failed to open for device: %s ret %d\n",
		       __func__, dev->name, ret);
		return -1;
	}
	return 0;
}

static int ltq_gsw_pmac_init(void)
{
	u8 i = 0, j = 0, k = 0;
	GSW_PMAC_Eg_Cfg_t eg_cfg;
	GSW_PMAC_Ig_Cfg_t ig_cfg;
	GSW_CPU_PortCfg_t cpu_port_cfg;
	GSW_portCfg_t gsw_port_cfg;

	/* Do the GSW-L configuration */
	struct core_ops *ops = gsw_get_swcore_ops(0);

	if (!ops) {
		pr_err("%s: Open SWAPI device FAILED!\n", __func__);
		return -EIO;
	}

	memset((void *)&eg_cfg, 0x00, sizeof(eg_cfg));
	memset((void *)&ig_cfg, 0x00, sizeof(ig_cfg));
	memset((void *)&cpu_port_cfg, 0x00, sizeof(cpu_port_cfg));

	/**
	 * 1. GSWIP-L PMAC Egress Configuration Table
	 * Entry:
	 * Address: (i from 0 to 15, j from 0 to 3)
	 * Destination Port = 0, Traffic Class = i, Others field = j
	 * Value:
	 * DMA Channel = if i < 3? i: 3
	 * PMAC_Flag = 1
	 * Other fields = 0
	 */
	pr_info("PMAC_EG_CFG_SET for GSW-L\n");

	for (i = 0; i <= 15; i++) {
		for (j = 0; j <= 3; j++) {
			memset((void *)&eg_cfg, 0x00, sizeof(eg_cfg));
#if defined(SINGLE_RX_CH0_ONLY) && SINGLE_RX_CH0_ONLY
			eg_cfg.nRxDmaChanId = 0;
#else
			eg_cfg.nRxDmaChanId = (i < 3) ? i : 3;
#endif
			eg_cfg.bRemL2Hdr = 0;
			eg_cfg.numBytesRem = 0;
			eg_cfg.bFcsEna = 0;
			eg_cfg.bPmacEna = 1;
			eg_cfg.nResDW1 = 0;
			eg_cfg.nRes1DW0 = 0;
			eg_cfg.nRes2DW0 = 0;
			eg_cfg.nDestPortId = 0;
			eg_cfg.nTrafficClass = i;
			eg_cfg.nFlowIDMsb = j;
			eg_cfg.bDecFlag = 0;
			eg_cfg.bEncFlag = 0;
			eg_cfg.bMpe1Flag = 0;
			eg_cfg.bMpe2Flag = 0;
			eg_cfg.bTCEnable = 1;
			ops->gsw_pmac_ops.Pmac_Eg_CfgSet(ops, &eg_cfg);
		}
	}

	/**
	 * 2. GSWIP-L PMAC Ingress Configuration Table
	 * Entry:
	 * Address: (i from 0 to 15)
	 * DMA Channel = i
	 * Value:
	 * PMAC_Flag = 1
	 * SPPID_Mode = 1
	 * SUBID_Mode = 1
	 * CLASSEN_Mode = 1
	 * CLASS_Mode = 1
	 * PMAPEN_Mode = 1
	 * PMAP_Mode = 1
	 * TCPERR_DP = 1
	 * DF_PMAC_HD.PMAP_EN = 1, DF_PMAC_HD.PMAP = 1 << i[2:0],
	 * DF_PMAC_HD.CLASS_EN = 1, DF_PMAC_HD.CLASS = i[3]*2 + 1
	 * Other fields = 0
	 */
	pr_info("PMAC_IG_CFG_SET for GSW-L\n");

	for (i = 0; i <= 15; i++) {
		memset((void *)&ig_cfg, 0x00, sizeof(ig_cfg));

		ig_cfg.nTxDmaChanId = i;
		ig_cfg.bErrPktsDisc = 1;
		ig_cfg.bPmapDefault = 1;
		ig_cfg.bPmapEna = 1;
		ig_cfg.bClassDefault = 1;
		ig_cfg.bClassEna = 1;
		ig_cfg.eSubId = 1;
		ig_cfg.bSpIdDefault = 1;
		ig_cfg.bPmacPresent = 1;
		ig_cfg.defPmacHdr[0] = 0;
		ig_cfg.defPmacHdr[1] = 0;
		ig_cfg.defPmacHdr[2] = ((((i & 8) >> 3) * 2) + 1); /* 1/3 */
		/* (1 << 7) | (1 << 4); */
		ig_cfg.defPmacHdr[3] = 0x90;
		ig_cfg.defPmacHdr[4] = 0;
		ig_cfg.defPmacHdr[5] = 0;
		/*if ( i >= 8)
		 *ig_cfg.defPmacHdr[6] = 1 << (i & 0x7);
		 *else
		 */
		ig_cfg.defPmacHdr[7] = 1 << (i & 0x7);
		ops->gsw_pmac_ops.Pmac_Ig_CfgSet(ops, &ig_cfg);
	}

	/* Enable the Ingress Special Tag */
	cpu_port_cfg.nPortId = 0;
	cpu_port_cfg.bCPU_PortValid = 0;
	cpu_port_cfg.bSpecialTagIngress = 1;
	cpu_port_cfg.bSpecialTagEgress = 0;
	cpu_port_cfg.bFcsCheck = 0;
	cpu_port_cfg.bFcsGenerate = 1;
	ops->gsw_common_ops.CPU_PortCfgSet(ops, &cpu_port_cfg);

#ifdef CONFIG_USE_EMULATOR
	/* Add some extra register writes for the 1.4.1 */
	gsw_reg_set_val(ops, 0x90f, 0x182);

	gsw_reg_set_val(ops, 0x91b, 0x182);

	gsw_reg_set_val(ops, 0x927, 0x182);

	gsw_reg_set_val(ops, 0x933, 0x182);

	gsw_reg_set_val(ops, 0x93f, 0x182);

#endif

	/* Enable the CPU port MAC address spoofing detection */
	gsw_reg_set_bit(ops, 0x480, 0x4000);

	/*PMAC control register 2 to disable LCHKS*/
	gsw_reg_clr_bit(ops, 0xd05, 0x3);

	/* Do the GSW-R configuration */
	ops = gsw_get_swcore_ops(1);

	if (!ops) {
		pr_err("%s: Open SWAPI device FAILED!\n", __func__);
		return -EIO;
	}

	/**
	 * 3. GSWIP-R PMAC Egress Configuration Table
	 * Entry:
	 * Address: (k i from 0 to 15,j from 0 to 63)
	 * Destination Port = k, Others field = j
	 * Value:
	 * DMA Channel = if k <= 13? 0: k
	 * PMAC_Flag = if k ==15 or k < 7? 1: 0
	 * FCS_Mode = if k ==13? PTM_Bonding_Mode:0
	 * L2_HD_RM = if k == 12 or k == 14? 1:0
	 * L2_HD_Bytes = if k == 14? CAPWAP_HD_BYTES:0
	 * Other fields = 0
	 */
	pr_info("PMAC_EG_CFG_SET for GSW-R\n");

	for (k = 0; k <= 15; k++) {
		for (i = 0; i <= 15; i++) {
			for (j = 0; j <= 3; j++) {
				memset((void *)&eg_cfg, 0x00, sizeof(eg_cfg));
#ifdef CONFIG_USE_EMULATOR
				eg_cfg.nRxDmaChanId = (k <= 14) ? 0 : 5;
#else

				if (ltq_get_soc_rev() == 2)
					eg_cfg.nRxDmaChanId = (k <= 13) ? 0 : 5;
				else
					eg_cfg.nRxDmaChanId = (k <= 13) ? 0 : k;

#endif
				/*((k == 15) || (k < 7)) ? 1 : 0; */
				eg_cfg.bPmacEna = 1;
				/*PTM_Bonding_Mode*/
				eg_cfg.bFcsEna = (k == 13) ? 1 : 0;
				eg_cfg.bRemL2Hdr = 0;
				/*CAPWAP_HD_BYTES*/
				eg_cfg.numBytesRem = (k == 14) ? 1 : 0;
				eg_cfg.nResDW1 = 0;
				eg_cfg.nRes1DW0 = 0;
				eg_cfg.nRes2DW0 = 0;
				eg_cfg.nDestPortId = k;
				eg_cfg.nTrafficClass = i;
				eg_cfg.bMpe1Flag = 0;
				eg_cfg.bMpe2Flag = 0;
				eg_cfg.bEncFlag = 0;
				eg_cfg.bDecFlag = 0;
				eg_cfg.nFlowIDMsb = j;
				eg_cfg.bTCEnable = 1;

				ops->gsw_pmac_ops.Pmac_Eg_CfgSet(ops, &eg_cfg);
			}
		}
	}

	/**
	 * 4. GSWIP-R PMAC Ingress Configuration Table
	 * Entry:
	 * Address: (i from 0 to 15)
	 * DMA Channel  = i
	 * Value:
	 * PMAC_Flag = if 5 < i < 15? 0: 1
	 * SPPID_Mode = 0
	 * SUBID_Mode = If 7 <= i <=8? 0: 1
	 * CLASSEN_Mode = 0
	 * CLASS_Mode = 0
	 * PMAPEN_Mode = if i < 4 or i == 15? 1:0
	 * PMAP_Mode = if i < 4 or i == 15? 1:0
	 * TCPERR_DP = 1
	 * DF_PMAC_HD.SPPID = i, DF_PMAC_HD.PMAP = if i == 15? 0x8000:0xFFFF
	 * DF_PMAC_HD.PMAC_EN = if i==15?1:0
	 * Other fields = 0
	 */
	pr_info("PMAC_IG_CFG_SET for GSW-R\n");

	for (i = 0; i <= 15; i++) {
		memset((void *)&ig_cfg, 0x00, sizeof(ig_cfg));

		ig_cfg.nTxDmaChanId = i;
		ig_cfg.bPmacPresent = 1; /*((i > 5) && (i < 15)) ? 0 : 1; */
		/*ig_cfg.bSpIdDefault  = 0; */
		/* For channel 15, use source port ID from default PMAC header*/
		ig_cfg.bSpIdDefault = (i == 15) ? 1 : 0;
		ig_cfg.eSubId = 0; /*((i == 6) || (i == 13)) ? 0 : 1;*/
		ig_cfg.bClassDefault = 0;
		ig_cfg.bClassEna = 0;
		ig_cfg.bErrPktsDisc = 1;

		ig_cfg.bPmapDefault = ((i < 4) || (i == 15)) ? 1 : 0;
		ig_cfg.bPmapEna = ((i < 4) || (i == 15)) ? 1 : 0;

		ig_cfg.defPmacHdr[0] = 0;
		ig_cfg.defPmacHdr[1] = 0;
		/*ig_cfg.defPmacHdr[2] = (i << 4); // Byte 2 (Bits 7:4)*/
		/* For channel 15, source port is 0. */
		ig_cfg.defPmacHdr[2] = (i == 15) ? 0 : (i << 4);
		ig_cfg.defPmacHdr[3] = (i == 15) ? 0x80 : 0; /*Byte 3 (Bit 7) */
		ig_cfg.defPmacHdr[4] = 0;
		ig_cfg.defPmacHdr[5] = 0;
		ig_cfg.defPmacHdr[6] = (i == 15) ? 0x80 : 0xFF;
		ig_cfg.defPmacHdr[7] = (i == 15) ? 0x00 : 0xFF;

		ops->gsw_pmac_ops.Pmac_Ig_CfgSet(ops, &ig_cfg);
	}

	/* Enable the Ingress Special Tag */
	for (k = 0; k < 15; k++) {
		memset((void *)&cpu_port_cfg, 0x00, sizeof(cpu_port_cfg));

		cpu_port_cfg.nPortId = k;

		/* get the current values first */
		ops->gsw_common_ops.CPU_PortCfgGet(ops, &cpu_port_cfg);

		/* change the required values and set it back */
		cpu_port_cfg.bCPU_PortValid = 0;
		cpu_port_cfg.bSpecialTagIngress = 1;
		cpu_port_cfg.bSpecialTagEgress = 0;
		cpu_port_cfg.bFcsCheck = 0;
		cpu_port_cfg.bFcsGenerate = 1;
		ops->gsw_common_ops.CPU_PortCfgSet(ops, &cpu_port_cfg);
	}

	/* Disable the Learning in the GSWIP-R */
	for (k = 0; k < 15; k++) {
		memset((void *)&gsw_port_cfg, 0x00, sizeof(gsw_port_cfg));

		gsw_port_cfg.nPortId = k;
		ops->gsw_common_ops.PortCfgGet(ops, &gsw_port_cfg);

		gsw_port_cfg.bLearning = 1;
		ops->gsw_common_ops.PortCfgSet(ops, &gsw_port_cfg);
	}

#ifdef CONFIG_USE_EMULATOR
	/* Add some extra register writes for the 1.4.1 */
	gsw_reg_set_val(ops, 0x903, 0x182);
#endif

	gsw_reg_set_val(ops, 0x454, 0x1);

	gsw_reg_set_val(ops, 0x455, 0x1);

	/*PMAC control register 2 to disable LCHKS*/
	gsw_reg_clr_bit(ops, 0xd05, 0x3);

	pr_info("\n\t GSW PMAC Init Done!!!\n");
	return 0;
}

static void
ltq_eth_drv_eth_addr_setup(struct net_device *dev, int port, int wan)
{
	int i = 0;
	u8 *macaddr_param = NULL;
	u8 values[ETH_ALEN] = { 0 };
	struct ltq_eth_priv *priv = netdev_priv(dev);

	if (is_valid_ether_addr(dev->dev_addr))
		return;

	if (wan)
		macaddr_param = macaddr1;
	else
		macaddr_param = macaddr0;

	if (macaddr_param) {
		mac_pton(macaddr_param, values);
		pr_debug("mac addr %s\n", macaddr_param);
	}

	/* read MAC address from the MAC table and put them into device */
	if (is_valid_ether_addr(values)) {
		pr_debug("using module param for port %d.\n", port);
		for (i = 0; i < ETH_ALEN; i++)
			dev->dev_addr[i] = values[i];
		if (multimac)
			dev->dev_addr[ETH_ALEN - 1] += port;
	} else {
		eth_hw_addr_random(dev);
		pr_debug("using random mac for port %d.\n", port);
	}
	if (priv->xgmac_id >= 0) {
		struct mac_ops *ops;

		ops = gsw_get_mac_ops(0, priv->xgmac_id);
		if (ops)
			ops->set_macaddr(ops, dev->dev_addr);
	}
}

#ifdef CONFIG_USERSPACE_LINK_NOTIFICATION
enum eth_cmd {
	ETH_C_UNSPEC,
	ETH_C_NOTIFY,
	__ETH_C_MAX,
};

#define ETH_C_MAX (__ETH_C_MAX - 1)

enum eth_attr {
	ETH_A_UNSPEC,
	ETH_A_LINK_NAME,
	ETH_A_LINK_STATUS,
	ETH_A_LINK_SPEED,
	__ETH_A_MAX,
};

#define ETH_LINK_A_MAX (__ETH_A_MAX - 1)

enum eth_multicast_groups {
	ETH_MCGRP,
};

/* VRX318 TC message multicast group */
static struct genl_multicast_group eth_grps[] = {
	[ETH_MCGRP] = {
		.name = "eth_mcgrp",
	},
};

/* VRX318 TC message genelink family */
static struct genl_family eth_gnl_family = {
	.id = GENL_ID_GENERATE, /* To generate an id for the family*/
	.hdrsize = 0,
	.name = "eth_drv_notify", /*family name, used by userspace application*/
	.version = 1, /*version number  */
	.maxattr = ETH_LINK_A_MAX,
	.mcgrps = eth_grps,
	.n_mcgrps = ARRAY_SIZE(eth_grps),
};

/* API definition for the driver to send TC notify messages to user application
 * using genetlink method.
 * pid: process id
 * tc_msg: 1-link is up,0- link is down
 * ln_no: interface name
 */
static int
ltq_eth_nl_msg_send(int pid, int link_status, int speed, char *if_name)
{
	struct sk_buff *skb;
	int ret;
	void *msg_head;

	skb = genlmsg_new(NLMSG_GOODSIZE, GFP_KERNEL);

	if (!skb)
		return -ENOMEM;

	/* create the message headers */
	msg_head = genlmsg_put(skb, 0, 0,
			       &eth_gnl_family, 0, ETH_C_NOTIFY);

	if (!msg_head) {
		ret = -ENOMEM;
		pr_err("create message header fail!\n");
		goto out;
	}

	nla_put_string(skb, ETH_A_LINK_NAME, if_name);

	if (link_status) {
		nla_put_string(skb, ETH_A_LINK_STATUS, "up");
		nla_put_u16(skb, ETH_A_LINK_SPEED, speed);
	} else {
		nla_put_string(skb, ETH_A_LINK_STATUS, "down");
	}
	genlmsg_end(skb, msg_head);
	ret = genlmsg_multicast(&eth_gnl_family, skb, pid,
				ETH_MCGRP, GFP_KERNEL);

	if (ret != 0 && ret != -ESRCH) {
		pr_err("failed to send out the multicast message:ret = %d\n",
		       ret);
		goto out;
	}

	return 0;
out:
	return ret;
}

static int ltq_eth_genetlink_init(void)
{
	int ret;

	/*register new family*/
	ret = genl_register_family(&eth_gnl_family);

	if (ret != 0) {
		pr_err("Family registeration fail:%s\n",
		       eth_gnl_family.name);
	}

	return ret;
}

static void ltq_eth_genetlink_exit(void)
{
	int ret;

	/*unregister the family*/
	ret = genl_unregister_family(&eth_gnl_family);

	if (ret != 0)
		pr_err("unregister Genetlink family %i\n", ret);
}
#endif

/* This function is event handler for net_device change notify */
static int phy_netdevice_event(struct notifier_block *nb, unsigned long action,
			       void *ptr)
{
	struct ltq_eth_priv *priv = NULL;
	struct net_device *dev = NULL;

	if (!ptr)
		return 0;

	dev = (struct net_device *)ptr;

	switch (action) {
	case NETDEV_CHANGENAME: {
		priv = netdev_priv(dev);

		if (priv && priv->wan) {
			pr_debug("\nUpdate WAN iface from [%s] to [%s]\n",
				 wan_iface, dev->name);
			memcpy(wan_iface, dev->name, IFNAMSIZ);
		}
	}
	}

	return NOTIFY_OK;
}

#if IS_ENABLED(CONFIG_AVM_PA)
static void ltq_pa_dev_transmit(void *arg, struct sk_buff *skb)
{
	int rc;

	skb->dev = arg;
	rc = dev_queue_xmit(skb);
	if ((rc & NETDEV_TX_MASK) != NETDEV_TX_OK && net_ratelimit()) {
		pr_err("[%s] pa_dev_transmit(%s) %d",
		       __func__, ((struct net_device *)arg)->name, rc);
	}
}

static int setup_avm_pa_eth_dev(struct net_device *dev)
{
	/* register avm_pa pid and set hw info */
	struct avm_pa_pid_cfg cfg = {0};
	struct avm_pa_vpid_cfg vcfg = {0};
	struct avm_pa_pid_hwinfo hwinfo = {0};
	int err;

	pr_err("[%s] dev %s\n", __func__, dev->name);

	strncpy(cfg.name, dev->name, sizeof(cfg.name));
	cfg.tx_func = &ltq_pa_dev_transmit;
	cfg.tx_arg = dev;
	cfg.default_mtu = 1500;
	cfg.framing = avm_pa_framing_ether;
	err = avm_pa_dev_pid_register(AVM_PA_DEVINFO(dev), &cfg);
	if (err < 0)
		return err;

	strncpy(vcfg.name, dev->name, sizeof(vcfg.name));
	vcfg.v4_mtu = 1500;
	vcfg.v6_mtu = 1500;
	err = avm_pa_dev_vpid_register(AVM_PA_DEVINFO(dev), &vcfg);
	if (err < 0)
		goto undo;

	hwinfo.ppa.netdev = dev;
	err = avm_pa_pid_set_hwinfo(AVM_PA_DEVINFO(dev)->pid_handle, &hwinfo);
undo:
	if (err < 0)
		avm_pa_dev_unregister_sync(AVM_PA_DEVINFO(dev));

	return err;
}
#endif

static void
xrx500_of_port(struct net_device *dev, struct device_node *port)
{
	const __be32 *addr, *id = of_get_property(port, "reg", NULL);
	struct resource irqres;
	struct ltq_eth_priv *priv = netdev_priv(dev);
	struct xrx500_port *p = &priv->port[priv->num_port];

	if (!id)
		return;

	memset(p, 0, sizeof(struct xrx500_port));
	p->phy_node = of_parse_phandle(port, "phy-handle", 0);
	p->phy_if = of_get_phy_mode(port);

	if (!p->phy_node) {
		/* check if a fixed-link is defined in device-tree */
		if (of_phy_is_fixed_link(port)) {
			struct phy_device *phy;

			pr_debug("fixed-link detected\n");
			if (of_phy_register_fixed_link(port)) {
				pr_err("cannot register fixed PHY\n");
				return;
			}
			p->phy_node = of_node_get(port);
			phy = of_phy_connect(dev,
					     p->phy_node, NULL, 0,
					     p->phy_if);
			if (!phy)
				pr_err("can not get phy dev for fixed link\n");
			p->phydev = phy;
			netif_carrier_on(dev);
		} else {
			pr_info("no interface attached to this interface!\n");
			return;
		}
	} else {
		addr = of_get_property(p->phy_node, "reg", NULL);
		if (!addr) {
			pr_info("no real PHY attached to this interface!\n");
			return;
		}
		p->phy_addr = *addr;
		if (of_irq_to_resource_table(port, &irqres, 1) == 1)
			p->irq_num = irqres.start;
		else
			pr_info("couldn't get irq number for gphy !!\n");
	}

	p->num = *id;
	priv->num_port++;

	/* is this port a wan port ? */
	if (priv->wan)
		priv->hw->wan_map |= BIT(p->num);

	priv->port_map |= BIT(p->num);

	/* store the port id in the hw struct so we can map ports -> devices */
	priv->hw->port_map[p->num] = priv->hw->num_devs;
}

static int xrx500_of_iface(struct xrx500_hw *hw, struct device_node *iface,
			   struct platform_device *pdev)
{
	struct ltq_eth_priv *priv;
	struct device_node *port;
	const __be32 *wan;
	u32 dp_dev_port_param, dp_port_id_param, xgmac_id_param;
	dp_cb_t cb = { 0 };
	u32 dp_port_id = 0;
	char name[16];
	u32 eee_adv;
	const char *dtname, *dtmacname, *macaddr;
	int ret;

	/* alloc the network device */
	hw->devs[hw->num_devs] = alloc_etherdev(sizeof(struct ltq_eth_priv));

	if (!hw->devs[hw->num_devs]) {
		pr_debug("allocated failed for interface %d\n",
			 hw->num_devs);
		return -ENOMEM;
	}

	priv = netdev_priv(hw->devs[hw->num_devs]);

	/* is this the wan interface ? */
	wan = of_get_property(iface, "lantiq,wan", NULL);

	if (wan && (*wan == 1))
		priv->wan = 1;
	else
		priv->wan = 0;

	dtname = of_get_property(iface, "avm,name", NULL);
	if (!dtname) {
		/* [TODO] avoid name reuse */
		pr_err("No interface name set, falling back to UGW defaults.\n");
		if (priv->wan)
			snprintf(name, sizeof(name), "eth1");
		else
			snprintf(name, sizeof(name), "eth0_%d", hw->num_devs);
	} else {
		strlcpy(name, dtname, sizeof(name));
	}

	ret = of_property_read_u32(iface, "intel,dp-dev-port",
				   &dp_dev_port_param);
	if (ret < 0) {
		pr_info("ERROR : Property intel,dp-dev-port not read from DT for if %s\n",
			name);
		return ret;
	}

	ret = of_property_read_u32(iface, "intel,dp-port-id",
				   &dp_port_id_param);
	if (ret < 0) {
		pr_info("ERROR : Property intel,dp-port-id not read from DT for if %s\n",
			name);
		return ret;
	}

	ret = of_property_read_u32(iface, "intel,xgmac-id",
				   &xgmac_id_param);
	if (ret < 0) {
		pr_debug("Property intel,xgmac-id not exist for if %s\n",
			 name);
		priv->xgmac_id = -1;
	} else {
		priv->xgmac_id = xgmac_id_param;
	}

	dtmacname = of_get_property(iface, "avm,macname", NULL);
	if (dtmacname) {
		macaddr = prom_getenv(dtmacname);
		if (macaddr) {
			mac_pton(macaddr, hw->devs[hw->num_devs]->dev_addr);
		}
	}

	ret = of_property_read_u32(iface, "avm,eee-adv", &eee_adv);
	if (ret < 0)
		eee_adv = 0;

	priv->eee_initial_adv = eee_adv;
	priv->eee_enabled = !!eee_adv;

	strcpy(hw->devs[hw->num_devs]->name, name);
	hw->devs[hw->num_devs]->netdev_ops = &ltq_eth_drv_ops;
	hw->devs[hw->num_devs]->watchdog_timeo = LTQ_TX_TIMEOUT;
	hw->devs[hw->num_devs]->needed_headroom = sizeof(ltq_pmac_header_t);
	SET_NETDEV_DEV(hw->devs[hw->num_devs], &pdev->dev);

	/* setup our private data */
	priv->hw = hw;
	priv->id = hw->num_devs;
	spin_lock_init(&priv->lock);

	priv->owner = &g_ltq_eth_module[hw->num_devs];
	sprintf(priv->owner->name, "module%02d", priv->id);

	if (priv->wan)
		dp_port_id = dp_alloc_port(priv->owner, hw->devs[hw->num_devs],
					   dp_dev_port_param, dp_port_id_param,
					   NULL, DP_F_FAST_ETH_WAN);
	else
		dp_port_id = dp_alloc_port(priv->owner, hw->devs[hw->num_devs],
					   dp_dev_port_param, dp_port_id_param,
					   NULL, DP_F_FAST_ETH_LAN);

	if (dp_port_id == DP_FAILURE) {
		pr_err("dp_alloc_port failed for %s with port_id %d\n",
		       hw->devs[hw->num_devs]->name, priv->id + 1);
		return -ENODEV;
	}

	priv->dp_port_id = dp_port_id;
	cb.stop_fn = (dp_stop_tx_fn_t)dp_fp_stop_tx;
	cb.restart_fn = (dp_restart_tx_fn_t)dp_fp_restart_tx;
	cb.rx_fn = (dp_rx_fn_t)dp_fp_rx;
#ifdef CONFIG_XRX500_ETH_DRV_COC_SUPPORT
	cb.dp_coc_confirm_stat_fn = (dp_coc_confirm_stat)dp_fp_coc_confirm;
#endif

	if (dp_register_dev(priv->owner, dp_port_id, &cb, 0) != DP_SUCCESS) {
		pr_err("dp_register_dev failed for %s\n and port_id %d",
		       hw->devs[hw->num_devs]->name, dp_port_id);
		dp_alloc_port(priv->owner, hw->devs[hw->num_devs],
			      dp_dev_port_param, dp_port_id_param,
			      NULL, DP_F_DEREGISTER);
		return -ENODEV;
	}

	/* load the ports that are part of the interface */
	for_each_child_of_node (iface, port) {
		if (of_device_is_compatible(port, "lantiq,xrx500-pdi-port"))
			xrx500_of_port(hw->devs[hw->num_devs], port);
	}

	hw->devs[hw->num_devs]->features = NETIF_F_SG;
#ifdef CONFIG_LTQ_TOE_DRIVER
	if (ltq_tso_available())
		hw->devs[hw->num_devs]->features |=
			NETIF_F_TSO | NETIF_F_HW_CSUM;
	hw->devs[hw->num_devs]->gso_max_size = GSO_MAX_SIZE;
#endif
	hw->devs[hw->num_devs]->hw_features = hw->devs[hw->num_devs]->features;
	hw->devs[hw->num_devs]->vlan_features = hw->devs[hw->num_devs]->features;

	if (g_soc_data.hw_checksum) {
		hw->devs[hw->num_devs]->features |= NETIF_F_HW_CSUM;
		hw->devs[hw->num_devs]->hw_features |= NETIF_F_HW_CSUM;
		pr_debug("%s: hw csum offload is enabled!\n", __func__);
	} else {
		pr_debug("%s: hw csum offload is disable!\n", __func__);
	}

	if (g_rx_csum_offload) {
		pr_info("%s: rx csum offload is enabled !\n", __func__);
		hw->devs[hw->num_devs]->features |= NETIF_F_RXCSUM;
		hw->devs[hw->num_devs]->hw_features |= NETIF_F_RXCSUM;
	} else {
		pr_info("%s: rx csum offload is disabled !\n", __func__);
	}
	ltq_eth_drv_eth_addr_setup(hw->devs[hw->num_devs], priv->id, priv->wan);

	/* register the actual device */
	if (!register_netdev(hw->devs[hw->num_devs])) {
#if IS_ENABLED(CONFIG_AVM_PA)
		ret = setup_avm_pa_eth_dev(hw->devs[hw->num_devs]);
		if (ret < 0)
			pr_err("register avmpa_dev failed: %d\n", ret);
#endif
		hw->num_devs++;
	}

	return 0;
}

static int xrx500_mmd_wr(struct mii_bus *bus, int addr, int reg, int devad, u16 val)
{
	GSW_MMD_data_t mmd_data;
	struct core_ops *ops = bus->priv;

	pr_debug("%s called with phy addr:%d and reg: %x and val: %x\n",
		 __func__, addr, reg, val);

	memset((void *)&mmd_data, 0x00, sizeof(mmd_data));
	mmd_data.nAddressDev = addr;
	mmd_data.nAddressReg = (devad << 16) | reg;
	mmd_data.nData = val;
	ops->gsw_common_ops.MmdDataWrite(ops, &mmd_data);

	return 0;
}

static int xrx500_mmd_rd(struct mii_bus *bus, int addr, int reg, int devad)
{
	GSW_MMD_data_t mmd_data;
	struct core_ops *ops = bus->priv;

	pr_debug("%s called with phy addr:%d and reg: %x\n",
		 __func__, addr, reg);

	memset((void *)&mmd_data, 0x00, sizeof(mmd_data));
	mmd_data.nAddressDev = addr;
	mmd_data.nAddressReg = (devad << 16) | reg;
	ops->gsw_common_ops.MmdDataRead(ops, &mmd_data);

	pr_debug("returing: %x\n", mmd_data.nData);
	return mmd_data.nData;
}

static int xrx500_mdio_wr(struct mii_bus *bus, int addr, int reg, u16 val)
{
	GSW_MDIO_data_t mmd_data;
	struct core_ops *ops = bus->priv;

	pr_debug("%s called with phy addr:%d and reg: %x and val: %x\n",
		 __func__, addr, reg, val);

	memset((void *)&mmd_data, 0x00, sizeof(mmd_data));
	mmd_data.nAddressDev = addr;
	mmd_data.nAddressReg = reg;
	mmd_data.nData = val;
	ops->gsw_common_ops.MDIO_DataWrite(ops, &mmd_data);

	return 0;
}

static int xrx500_mdio_rd(struct mii_bus *bus, int addr, int reg)
{
	GSW_MDIO_data_t mmd_data;
	struct core_ops *ops = bus->priv;

	pr_debug("%s called with phy addr:%d and reg: %x\n",
		 __func__, addr, reg);

	memset((void *)&mmd_data, 0x00, sizeof(mmd_data));
	mmd_data.nAddressDev = addr;
	mmd_data.nAddressReg = reg;
	ops->gsw_common_ops.MDIO_DataRead(ops, &mmd_data);

	pr_debug("returing: %x\n", mmd_data.nData);
	return mmd_data.nData;
}

static int xrx500_of_mdio(struct mii_bus **busp, int devid, struct device_node *np)
{
	struct mii_bus *bus = mdiobus_alloc();

	if (!bus)
		return -ENOMEM;

	bus->read = xrx500_mdio_rd;
	bus->write = xrx500_mdio_wr;
	bus->mmd_read = xrx500_mmd_rd;
	bus->mmd_write = xrx500_mmd_wr;
	bus->name = "lantiq,xrx500-mdio";
	snprintf(bus->id, MII_BUS_ID_SIZE, "%x", devid);
	bus->priv = gsw_get_swcore_ops(devid);

	if (!bus->priv) {
		pr_err("%s: Open SWAPI device FAILED!\n", __func__);
		return -EIO;
	}

	pr_info("registering one of MII bus\n");

	if (of_mdiobus_register(bus, np)) {
		mdiobus_free(bus);
		return -ENXIO;
	}

	*busp = bus;

	return 0;
}

#ifdef SW_POLLING
static void xrx500_gmac_update(struct xrx500_port *port)
{
	u16 phyaddr = port->phydev->addr & MDIO_PHY_ADDR_MASK;

	if (port->phydev->link)
		phyaddr |= MDIO_PHY_LINK_UP;
	else
		phyaddr |= MDIO_PHY_LINK_DOWN;

	if (port->phydev->duplex == DUPLEX_FULL)
		phyaddr |= MDIO_PHY_FDUP_EN;
	else
		phyaddr |= MDIO_PHY_FDUP_DIS;

	udelay(1);
}
#else
static void xrx500_gmac_update(struct xrx500_port *port)
{
}
#endif
static void xrx500_mdio_link(struct net_device *dev)
{
	struct ltq_eth_priv *priv;
	int i;

	priv = netdev_priv(dev);
	pr_debug("%s called..\n", __func__);

	for (i = 0; i < priv->num_port; i++) {
		if (!priv->port[i].phydev)
			continue;

		if (priv->port[i].link != priv->port[i].phydev->link) {
			xrx500_gmac_update(&priv->port[i]);
			priv->port[i].link = priv->port[i].phydev->link;
			/* netdev_info(dev, "port %d %s link\n",
			 *	priv->port[i].num,
			 *	(priv->port[i].link)?("got"):("lost"));
			 */
#ifdef CONFIG_USERSPACE_LINK_NOTIFICATION
			ltq_eth_nl_msg_send(0, priv->port[i].link,
					    priv->port[i].phydev->speed,
					    dev->name);
#endif
			phy_print_status(priv->port[i].phydev);
		}
	}
}

static int xrx500_mdio_probe(struct net_device *dev, struct xrx500_port *port)
{
	struct phy_device *phydev = NULL;
	struct ltq_eth_priv *priv = NULL;
	struct mii_bus *bus = NULL;
	struct ethtool_eee eee_initial = {};

	priv = netdev_priv(dev);

	if (priv->wan)
		bus = priv->hw->mii_bus_pae;
	else
		bus = priv->hw->mii_bus;

	if (!bus) {
		pr_info("No mdio bus defined for this port!\n");
		return -ENODEV;
	}

	phydev = mdiobus_get_phy(bus, port->phy_addr);
	if (!phydev) {
		netdev_err(dev, "no PHY found\n");
		return -ENODEV;
	}

	pr_info("trying to connect: %s to device: %s with irq: %d\n",
		dev->name, phydev_name(phydev), port->irq_num);
	phydev->irq = port->irq_num;
	phydev = phy_connect(dev, phydev_name(phydev), &xrx500_mdio_link,
			     port->phy_if);

	if (IS_ERR(phydev)) {
		netdev_err(dev, "Could not attach to PHY\n");
		return PTR_ERR(phydev);
	}

	phydev->supported &= (SUPPORTED_10baseT_Half | SUPPORTED_10baseT_Full | SUPPORTED_100baseT_Half | SUPPORTED_100baseT_Full | SUPPORTED_1000baseT_Half | SUPPORTED_1000baseT_Full | SUPPORTED_Autoneg | SUPPORTED_MII | SUPPORTED_TP);
	phydev->advertising = phydev->supported;
	port->phydev = phydev;

	eee_initial.advertised = priv->eee_initial_adv;

	pr_info("%s: attached PHY [%s] (phy_addr=%s, irq=%d)\n",
		dev->name, phydev->drv->name,
		phydev_name(phydev), phydev->irq);

	phy_ethtool_set_eee(phydev, &eee_initial);
	phy_read_status(phydev);
	phy_start_aneg(phydev);
	return 0;
}

#ifdef CONFIG_XRX500_ETH_DRV_COC_SUPPORT
/* switch api related routines */
/*Base address GSWIP-L */
static void __iomem *ltq_eth_gswl_base = (void __iomem *)KSEG1ADDR(0x1c000000);
/* Base address GSWIP-R */
static void __iomem *ltq_eth_gswr_base = (void __iomem *)KSEG1ADDR(0x1a000000);

/** read the gswitch register */
static void ltq_eth_gsw_r32(void __iomem *gsw_base, short offset, short shift,
			    short size, u32 *value)
{
	u32 rvalue, mask;

	rvalue = ltq_r32(gsw_base + (offset * 4));
	mask = (1 << size) - 1;
	rvalue = (rvalue >> shift);
	*value = (rvalue & mask);
}

/** read and update the GSWIP register */
static void ltq_eth_gsw_w32(void __iomem *gsw_base, short offset, short shift,
			    short size, u32 value)
{
	u32 rvalue, mask;

	rvalue = ltq_r32(gsw_base + (offset * 4));
	mask = (1 << size) - 1;
	mask = (mask << shift);
	value = ((value << shift) & mask);
	value = ((rvalue & ~mask) | value);
	ltq_w32(value, (gsw_base + (offset * 4)));
}

static int32_t dp_fp_coc_confirm(enum ltq_cpufreq_state new_state,
				 enum ltq_cpufreq_state old_state,
				 u32 flags)
{
	GSW_register_t reg_cfg;
	int ret;

	if ((new_state != g_ltq_eth_drv_coc_state) &&
	    (new_state != LTQ_CPUFREQ_PS_D0)) {
		pr_debug("%s: CoC state changed from: %d to : %d enable irq!\n",
			 __func__, g_ltq_eth_drv_coc_state, new_state);

		/* Do the GSW-L configuration */
		struct core_ops *ops = gsw_get_swcore_ops(0);

		if (!ops) {
			pr_err("%s: Open SWAPI device FAILED!\n", __func__);
			return -EIO;
		}

		/* Clear the BM interrupt */
		gsw_reg_set_val(ops, 0x55, 0x100);

		/* Enable the BM interrupt */
		gsw_reg_set_bit(ops, 0x14, 0x1);

		/* Enable the QCGN interrupt */
		gsw_reg_set_bit(ops, 0x54, 0x100);

		/* Do the PAE configuration */
		ops = gsw_get_swcore_ops(1);

		if (!ops) {
			pr_err("%s: Open SWAPI device FAILED!\n", __func__);
			return -EIO;
		}

		/* Clear all the pending metering interrupt */
		gsw_reg_set_val(ops, 0x489, 0x100);

		/* Enable the PCE interrupt */
		gsw_reg_set_bit(ops, 0x14, 0x2);

		/* Enable PCE interrupt for port 0 */
		gsw_reg_set_bit(ops, 0x465, 0x1);

		/* Enable metering interrupt inside PCE */
		gsw_reg_set_bit(ops, 0x488, 0x100);

	} else if (g_ltq_eth_drv_coc_state != new_state &&
		   new_state == LTQ_CPUFREQ_PS_D0) {
		pr_debug("%s: CoC state changed from : %d to : %d.\n",
			 __func__, g_ltq_eth_drv_coc_state, new_state);
		/*disable_irq(g_ltq_eth_gswl_irq);*/
	} else {
		pr_debug("duplicate call to confirm to new state: %d!\n",
			 new_state);
	}

	g_ltq_eth_drv_coc_state = new_state;
	return 0;
}

static void ltq_eth_gswl_tasklet(unsigned long ptr)
{
	GSW_register_t reg_cfg;
	GSW_API_HANDLE gswl;
	u32 intr_type, irq_status;
	int ret = 0;

	pr_debug("tasklet called..\n");
	/* Do the GSW-L configuration */
	struct core_ops *ops = gsw_get_swcore_ops(0);

	if (!ops) {
		pr_err("%s: Open SWAPI device FAILED!\n", __func__);
		return -EIO;
	}

	while (!bitmap_empty(g_ltq_eth_intr_type, LTQ_ETH_NUM_INTERRUPTS)) {
		intr_type = find_first_bit(g_ltq_eth_intr_type,
					   LTQ_ETH_NUM_INTERRUPTS);

		/* Read the status */
		switch (intr_type) {
		case 1:

			irq_status = gsw_reg_get_val(ops, 0x55);
			pr_debug("BM interrupt with status: %x!\n", irq_status);

			/* Egress Queue Congestion Status Change */
			if (irq_status & 0x100) {
				/* Mask the interrupt */
				gsw_reg_clr_bit(ops, 0x54, 0x100);

				/* Clear all the interrupts */
				gsw_reg_set_val(ops, 0x55, irq_status);

				/* Clear at the top level */
				gsw_reg_set_val(ops, 0x15, 1);

				/* There is some traffic, so call the dp library
				 * to change the power state to upscale
				 */
				dp_coc_new_stat_req(LTQ_CPUFREQ_PS_D0,
						    DP_COC_REQ_ETHERNET);

			} else {
				pr_err("non QCGN interrupts from BM are not handled !");
			}
			clear_bit(1, g_ltq_eth_intr_type);
			break;

		default:
			pr_err("%s: this irq is not handled..\n", __func__);
			break;
		}
	}
}

static irqreturn_t ltq_eth_gswl_isr(int irq, void *dev_id)
{
	u32 irq_status, irq_mask;

	ltq_eth_gsw_r32(ltq_eth_gswl_base, 0x15, 0, 16, &irq_status);

	/* Buffer Manager Interrupt */
	if (irq_status & 0x1) {
		/* Mask the BM Interrupt */
		ltq_eth_gsw_r32(ltq_eth_gswl_base, 0x14, 0, 16, &irq_mask);

		irq_mask &= ~0x1;
		ltq_eth_gsw_w32(ltq_eth_gswl_base, 0x14, 0, 16, irq_mask);

		set_bit(1, g_ltq_eth_intr_type);

		/* Schedule the tasklet */
		tasklet_schedule(&gswl_tasklet);
	} else {
		pr_err("%s: ..........: %x", __func__, irq_status);
	}

	return IRQ_HANDLED;
}

/* PAE interrupt related */

static irqreturn_t ltq_eth_pae_isr(int irq, void *dev_id)
{
	u32 irq_status, irq_mask;

	ltq_eth_gsw_r32(ltq_eth_gswr_base, 0x15, 0, 16, &irq_status);

	/* PCE interrupt */
	if (irq_status & 0x2) {
		ltq_eth_gsw_r32(ltq_eth_gswr_base, 0x14, 0, 16, &irq_mask);

		irq_mask &= ~0x2;
		ltq_eth_gsw_w32(ltq_eth_gswr_base, 0x14, 0, 16, irq_mask);

		set_bit(2, g_ltq_pae_intr_type);

		/* Schedule the tasklet */
		tasklet_schedule(&pae_tasklet);
	} else {
		pr_err("%s: ..........: %x", __func__, irq_status);
	}
	return IRQ_HANDLED;
}

static void ltq_eth_pae_tasklet(unsigned long ptr)
{
	GSW_register_t reg_cfg;
	GSW_API_HANDLE gswr;
	u32 intr_type, pce_irq_status;
	int ret;

	pr_debug("PAE tasklet called..\n");
	struct core_ops *ops = gsw_get_swcore_ops(1);

	if (!ops) {
		pr_err("%s: Open SWAPI device FAILED!\n", __func__);
		return -EIO;
	}

	while (!bitmap_empty(g_ltq_pae_intr_type, LTQ_ETH_NUM_INTERRUPTS)) {
		intr_type = find_first_bit(g_ltq_pae_intr_type,
					   LTQ_ETH_NUM_INTERRUPTS);

		switch (intr_type) {
		/* PCE interrupt */
		case 2:
			/* Find out the port */

			pr_debug("PCE isr_0: %x!\n",
				 gsw_reg_get_val(ops, 0x467));

			/* Support only for port 0 */
			if (gsw_reg_get_val(ops, 0x467) & 0x1) {
				pce_irq_status = gsw_reg_get_val(ops, 0x489);
				pr_debug("PCE isr_0 internal: %x!\n",
					 pce_irq_status);

				/* Metering based bp status change interrupt */
				if (pce_irq_status & 0x100) {
					/* Mask the interrupt */
					gsw_reg_clr_bit(ops, 0x488, 0x100);

					/* Clear all the interrupts */
					gsw_reg_set_val(ops, 0x467, 1);

					gsw_reg_set_val(ops, 0x489,
							pce_irq_status);

					/* There is some traffic, so call the
					 * datapth library to change the power
					 * state to upscale
					 */
					dp_coc_new_stat_req(LTQ_CPUFREQ_PS_D0,
							    DP_COC_REQ_ETHERNET);

				} else {
					pr_err("non metering intr not handled!");
					break;
				}

				clear_bit(2, g_ltq_pae_intr_type);
			} else {
				pr_err("%s: only port 0 irq is handled..\n",
				       __func__);
				return;
			}

			break;

		default:
			pr_err("%s: this irq is not handled..\n", __func__);
			break;
		}
	}
}
#endif

/* Initialization Ethernet module */
static int ltq_eth_drv_init(struct platform_device *pdev)
{
	int ret = 0;
	struct device_node *node = pdev->dev.of_node;
	struct device_node *mdio_np, *iface_np;
	struct mii_bus *bus;
#ifdef CONFIG_XRX500_ETH_DRV_COC_SUPPORT
	struct resource irqres[2];
#endif

	memset(g_ltq_eth_module, 0, sizeof(g_ltq_eth_module));

	ret = of_property_read_u32(node, "lantiq,eth-switch-mode",
				   &g_eth_switch_mode);

	if (ret < 0) {
		pr_info("couldn't get the eth switch mode from DT.");
		g_eth_switch_mode = 0;
		ret = 0;
	}

	if (g_eth_switch_mode > 0) {
		/* HW init of the Switch */
		if (ltq_gsw_pmac_init() < 0)
			pr_info("%s[%d]: switch PMAC init failed..\n",
				__func__, __LINE__);
	} else {
		pr_debug("Ethernet driver do not need to config the switch.\n");
	}

	ret = of_property_read_u32(node, "lantiq,eth-rx-csum-offload",
				   &g_rx_csum_offload);

	if (ret < 0) {
		dev_info(&pdev->dev,
			 "couldn't get the RX CSUM offload setting from DT.");
		dev_info(&pdev->dev, " by default its disabled !\n");
		ret = 0;
	}

	/* bring up the mdio bus */
	mdio_np = of_find_compatible_node(node, NULL,
					  "lantiq,xrx500-mdio");

	if (mdio_np) {
		if (xrx500_of_mdio(&xrx500_hw.mii_bus, 0, mdio_np))
			dev_err(&pdev->dev, "mdio probe failed\n");

		/* bring up the mdio bus for PAE */
		mdio_np = of_find_compatible_node(node, NULL,
						  "lantiq,xrx500-mdio-pae");

		if (mdio_np)
			if (xrx500_of_mdio(&xrx500_hw.mii_bus_pae, 1, mdio_np))
				dev_err(&pdev->dev, "mdio probe of PAE failed\n");

	} else {
		mdio_np = of_find_compatible_node(node, NULL,
						  "intel,falconmx-mdio");

		if (mdio_np)
			pr_debug("The mdio bus intel,falconmx-mdio found.\n");
		else
			pr_err("No MDIO bus defined!\n");
	}

	bus = xrx500_hw.mii_bus;
	if (bus)
		bus->parent = &pdev->dev;

	xrx500_hw.num_devs = 1;
	for_each_available_child_of_node (node, iface_np) {
		if (of_device_is_compatible(iface_np, "lantiq,xrx500-pdi")) {
			pr_debug("adding the interface: %d\n",
				 xrx500_hw.num_devs);

			if (!of_device_is_available(iface_np)) {
				pr_debug("device not available.\n");
				continue;
			}
			if (xrx500_hw.num_devs < NUM_ETH_INF) {
				xrx500_of_iface(&xrx500_hw, iface_np, pdev);
			} else {
				dev_err(&pdev->dev,
					"only %d interfaces allowed\n",
					NUM_ETH_INF);
			}
		}
	}

	if (!xrx500_hw.num_devs) {
		dev_err(&pdev->dev, "failed to load interfaces\n");
		return -ENOENT;
	}
#ifdef CONFIG_USERSPACE_LINK_NOTIFICATION
	/* Register the netlink notification */
	ltq_eth_genetlink_init();
#endif

#ifdef CONFIG_XRX500_ETH_DRV_COC_SUPPORT

	if (of_irq_to_resource_table(node, irqres, 2) == 2) {
		/* GSW-L interrupt */
		if (devm_request_irq(&pdev->dev, irqres[0].start,
				     ltq_eth_gswl_isr, 0,
				     "gswl_irq", NULL)) {
			pr_err("%s: failed to request gswl irq - ", __func__);
			goto err1;
		}

		tasklet_init(&gswl_tasklet, ltq_eth_gswl_tasklet, 0);
		bitmap_zero(g_ltq_eth_intr_type, LTQ_ETH_NUM_INTERRUPTS);
		/* Keep the interrupt disabled till the low power mode*/
		/*disable_irq(irqres.start);*/
		g_ltq_eth_gswl_irq = irqres[0].start;

		/* PAE interrupt */
		if (devm_request_irq(&pdev->dev, irqres[1].start,
				     ltq_eth_pae_isr, 0,
				     "pae_irq", NULL)) {
			pr_err("%s: failed to request pae irq - ",
			       __func__);
			goto err1;
		}

		tasklet_init(&pae_tasklet, ltq_eth_pae_tasklet, 0);
		bitmap_zero(g_ltq_pae_intr_type, LTQ_ETH_NUM_INTERRUPTS);
		g_ltq_pae_irq = irqres[1].start;
	} else {
		pr_err("%s: couldn't get irq from device tree for CoC.\n",
		       __func__);
	}

err1:
#endif
	pr_info("Lantiq ethernet driver for XRX500 init.\n");
	return 0;
}

static void ltq_eth_drv_exit(struct platform_device *pdev)
{
	int i, ret, j;

	for (i = 0; i < xrx500_hw.num_devs; i++) {
		struct ltq_eth_priv *priv;
		struct net_device *dev = xrx500_hw.devs[i];

		if (!dev)
			continue;

		netif_stop_queue(dev);

		priv = netdev_priv(dev);

		pr_debug("num port %d\n", priv->num_port);

		for (j = 0; j < priv->num_port; j++)
			if (priv->port[j].phydev) {
				struct device_node *device;

				device = priv->port[j].phy_node;
				pr_debug("phy_disconnect phydev(%d) 0x%p\n",
					 j, priv->port[j].phydev);
				phy_disconnect(priv->port[j].phydev);
				if (of_phy_is_fixed_link(device)) {
					of_phy_deregister_fixed_link(device);
					pr_debug("deregister the fixed link\n");
				}
			}

		pr_debug("ltq_eth_drv_exit i %d dev_id %d port_id %d for device %s\n",
			 i, priv->dev_port, priv->dp_port_id, dev->name);
		priv->dp_subif.subif = 0;
		priv->dp_subif.port_id = priv->dp_port_id;

		ret = dp_register_subif(priv->owner, dev, dev->name,
					&priv->dp_subif, DP_F_DEREGISTER);

		if (ret != DP_SUCCESS) {
			pr_err("%s: failed to call deregister subif: %s\n",
			       __func__, dev->name);
		}
		ret = dp_register_dev(priv->owner, priv->dp_port_id, NULL,
				      DP_F_DEREGISTER);
		if (ret != DP_SUCCESS) {
			pr_err("%s: failed to call deregister device: %s\n",
			       __func__, dev->name);
		}
		ret = dp_alloc_port(priv->owner, dev, priv->dev_port,
				    priv->dp_port_id, NULL, DP_F_DEREGISTER);
		if (ret != DP_SUCCESS) {
			pr_err("%s: failed to call dealloc for device : %s\n",
			       __func__, dev->name);
		}
		priv->dp_port_id = DP_FAILURE;
		unregister_netdev(dev);
		free_netdev(dev);
	}

	if (xrx500_hw.mii_bus) {
		mdiobus_unregister(xrx500_hw.mii_bus);
		mdiobus_free(xrx500_hw.mii_bus);
	}

	if (xrx500_hw.mii_bus_pae) {
		mdiobus_unregister(xrx500_hw.mii_bus_pae);
		mdiobus_free(xrx500_hw.mii_bus_pae);
	}

	memset(&xrx500_hw, 0, sizeof(xrx500_hw));
#ifdef CONFIG_USERSPACE_LINK_NOTIFICATION
	ltq_eth_genetlink_exit();
#endif
	pr_info("Lantiq ethernet driver for XRX500 remove.\n");
}

static int ltq_eth_drv_probe(struct platform_device *pdev)
{
	struct device *dev = &pdev->dev;
	struct ltq_net_soc_data *soc_data = NULL;

	soc_data = (struct ltq_net_soc_data *)of_device_get_match_data(dev);
	if (!soc_data) {
		pr_err("No data found for ltq eth drv!\n");
		return -EINVAL;
	}

	if (soc_data->need_defer && !is_xway_gphy_fw_loaded())
		return -EPROBE_DEFER;

	memcpy(&g_soc_data, soc_data, sizeof(*soc_data));
	/* Just do the init */
	ltq_eth_drv_init(pdev);
	register_netdevice_notifier(&netdevice_notifier);

	return 0;
}

static int ltq_eth_drv_remove(struct platform_device *pdev)
{
	/* Just do the exit */
	unregister_netdevice_notifier(&netdevice_notifier);
	ltq_eth_drv_exit(pdev);
	return 0;
}

static const struct ltq_net_soc_data xrx500_net_data = {
	.need_defer = true,
	.hw_checksum = true,
};

static const struct ltq_net_soc_data falconmx_net_data = {
	.need_defer = false,
	.hw_checksum = false,
};

static const struct of_device_id ltq_eth_drv_match[] = {
	{ .compatible = "lantiq,xrx500-eth", .data = &xrx500_net_data },
	{ .compatible = "lantiq,falconmx-eth", .data = &falconmx_net_data },
	{},
};
MODULE_DEVICE_TABLE(of, ltq_eth_drv_match);

static struct platform_driver ltq_eth_driver = {
	.probe = ltq_eth_drv_probe,
	.remove = ltq_eth_drv_remove,
	.driver = {
		.name = "xrx500-eth",
		.of_match_table = ltq_eth_drv_match,
		.owner = THIS_MODULE,
	},
};

enum avm_ethttype {
	LINUX_NETDEV_STATS_64,
	HW_RMON_STATS
};

struct avm_etht_stat {
	char stat_string[ETH_GSTRING_LEN];
	enum avm_ethttype type;
	size_t sizeof_stat;
	size_t stat_offset;
};

#define LINUX_NETDEV_STAT_64(str, m)                                       \
	{                                                                  \
		.stat_string = str,                                        \
		.type = LINUX_NETDEV_STATS_64,                             \
		.sizeof_stat = sizeof(((struct rtnl_link_stats64 *)0)->m), \
		.stat_offset = offsetof(struct rtnl_link_stats64, m)       \
	}

#define HW_RMON_STAT(str, m)                                          \
	{                                                             \
		.stat_string = str,                                   \
		.type = HW_RMON_STATS,                                \
		.sizeof_stat = sizeof(((GSW_RMON_Port_cnt_t *)0)->m), \
		.stat_offset = offsetof(GSW_RMON_Port_cnt_t, m)       \
	}

static const struct avm_etht_stat avm_etht_stats[] = {
	LINUX_NETDEV_STAT_64("netif_rx_pkts_good", rx_packets),
	LINUX_NETDEV_STAT_64("netif_tx_pkts_good", tx_packets),

	LINUX_NETDEV_STAT_64("netif_rx_bytes_good", rx_bytes),
	LINUX_NETDEV_STAT_64("netif_tx_bytes_good", tx_bytes),

	LINUX_NETDEV_STAT_64("netif_rx_pkts_dropped", rx_dropped),
	LINUX_NETDEV_STAT_64("netif_tx_pkts_dropped", tx_dropped),

	LINUX_NETDEV_STAT_64("netif_rx_pkts_error", rx_errors),
	LINUX_NETDEV_STAT_64("netif_tx_pkts_error", tx_errors),

	HW_RMON_STAT("mac_rx_pkts_good", nRxGoodPkts),
	HW_RMON_STAT("mac_tx_pkts_good", nTxGoodPkts),

	HW_RMON_STAT("mac_rx_bytes_good", nRxGoodBytes),
	HW_RMON_STAT("mac_tx_bytes_good", nTxGoodBytes),

	HW_RMON_STAT("mac_rx_pkts_pause", nRxGoodPausePkts),
	HW_RMON_STAT("mac_tx_pkts_pause", nTxPauseCount),

	HW_RMON_STAT("mac_rx_pkts_dropped", nRxDroppedPkts),
	HW_RMON_STAT("mac_tx_pkts_dropped", nTxDroppedPkts),

	HW_RMON_STAT("mac_rx_bytes_error", nRxBadBytes),
};

#define AVM_ETHTOOL_STATS_LEN ARRAY_SIZE(avm_etht_stats)

static int avmnet_get_sset_count(struct net_device *dev,
				 int string_set)
{
	switch (string_set) {
	case ETH_SS_STATS:
		return AVM_ETHTOOL_STATS_LEN;
	default:
		return -EINVAL;
	}
}

static void avmnet_get_strings(struct net_device *dev,
			       u32 stringset, u8 *data)
{
	int i;

	switch (stringset) {
	case ETH_SS_STATS:
		for (i = 0; i < AVM_ETHTOOL_STATS_LEN; i++) {
			memcpy(data + i * ETH_GSTRING_LEN,
			       avm_etht_stats[i].stat_string,
			       ETH_GSTRING_LEN);
		}
		break;
	}
}

static void avmnet_get_ethtool_stats(struct net_device *dev,
				     struct ethtool_stats *stats, u64 *data)
{
	char *p = NULL;
	struct core_ops *ops;
	struct rtnl_link_stats64 netdev_stat64_store = { 0 };
	GSW_RMON_Port_cnt_t rmon_stats = { 0 };
	int rmon_got;
	int i;
	struct ltq_eth_priv *priv = netdev_priv(dev);

	if (dev->netdev_ops->ndo_get_stats64) {
		dev->netdev_ops->ndo_get_stats64(dev, &netdev_stat64_store);
	}

	rmon_stats.nPortId = priv->dp_port_id;

	ops = gsw_get_swcore_ops(!!(priv->wan));
	if (!ops) {
		pr_err("Open GSWIP failed\n");
		return;
	}

	rmon_got = ops->gsw_rmon_ops.RMON_Port_Get(ops, &rmon_stats);
	if (rmon_got != GSW_statusOk) {
		pr_err("RMON_PORT_GET failed for Port %d\n", rmon_stats.nPortId);
		return;
	}

	for (i = 0; i < AVM_ETHTOOL_STATS_LEN; i++) {
		switch (avm_etht_stats[i].type) {
		case LINUX_NETDEV_STATS_64:
			p = ((char *)&netdev_stat64_store) +
			    avm_etht_stats[i].stat_offset;
			break;

		case HW_RMON_STATS:
			p = ((char *)&rmon_stats) +
			    avm_etht_stats[i].stat_offset;
			break;
		default:
			dev_err(&dev->dev, "%s: stats have unknown type %d\n",
				__func__, avm_etht_stats[i].type);
			return;
		}

		data[i] = (avm_etht_stats[i].sizeof_stat ==
			   sizeof(u64)) ?
				  *(u64 *)p :
				  *(u32 *)p;
	}
}

module_platform_driver(ltq_eth_driver);

module_param(macaddr0, charp, 0644);
module_param(macaddr1, charp, 0644);
module_param(multimac, int, 0644);
MODULE_PARM_DESC(macaddr0, "eth0 MAC address");
MODULE_PARM_DESC(macaddr1, "eth1 MAC address");
MODULE_PARM_DESC(multimac, "Multi MAC address for ports");

MODULE_DESCRIPTION("Intel ethernet driver");
MODULE_LICENSE("GPL v2");
MODULE_VERSION(DRV_MODULE_VERSION);
